import json
import logging
import os
import tempfile

from pymerkle import MerkleTree
from pymerkle.proof import Proof
from pymerkle import validateProof

from pruvi.exceptions import NoPartsException
from pruvi.hashing import hash_bytes

class BaseSplitter(object):
    """Base Splitter object

    Attributes:
        _parts (list): List of elements to proof.
        hash_type (str): The hash type to be used.
        tree: The Merkle Tree created.
        proofs (list): List of proofs.
    """
    def __init__(self, parts=[], hash_type="sha3_512"):
        """Constructor

        Args:
            parts (list): List of elements to proof.
            hash_type (str): The hash type to be used.
        """
        self._parts = []
        self.set_parts(parts)
        self.hash_type = hash_type
        self.tree = None
        self.proofs = []

    def _export_proofs(self, output_file):
        """Export all proofs to a file

        Args:
            output_file (str): Output file.
        """
        info = {
            "merkle_root": self.tree.rootHash.decode(),
            "proofs": self.proofs
        }

        with open(output_file, "w") as f:
            f.write(json.dumps(info, indent=2))

    def _export_tree(self, output_file):
        """Export tree

        Args:
            output_file (str): Output file.
        """
        self.tree.export(f"{output_file}")

    def _write_parts(self, output_folder=tempfile.gettempdir()):
        """Writes a file for each part

        Args:
            output_folder (str): The folder where the parts will be written.
        """
        if not os.path.exists(output_folder):
            logging.info(f"Creating parts folder at '{output_folder}'...")
            os.makedirs(output_folder)

        ext = os.path.splitext(self.file_path)[-1]

        for i, part in enumerate(self._parts):
            output_file = os.path.join(
                output_folder,
                f"part-{i+1}{ext}"
            )

            logging.info(f"Writing part {i+1} at '{output_file}'...")
            with open(output_file, "wb") as f:
                f.write(self._parts[i])

            output_file_proof = output_file.replace(ext, "-proof.json")
            logging.info(f"Writing proof for part {i+1} at '{output_file_proof}'...")
            with open(output_file_proof, "w") as f:
                f.write(json.dumps(self.proofs[i], indent=2))

    def create_tree(self):
        """Create the tree

        Args:
            output_folder (str): The folder where the parts will be written.
        """
        self.tree = MerkleTree(hash_type=self.hash_type, security=False, raw_bytes=True)

        # Creating tree
        for i, f in enumerate(self.get_parts()):
            x = self.tree.update(f)

        # Creating proofs
        for i, f in enumerate(self._parts):
            self.proofs.append(self.tree.auditProof(i).serialize())

    def export(self, output_folder):
        """Export the splitted information

        Args:
            output_folder (str): Output folder.
        """
        if not os.path.exists(output_folder):
            logging.info(f"Creating export folder at '{output_folder}'...")
            os.makedirs(output_folder)

        tree_file = os.path.join(output_folder, "tree.json")
        logging.info(f"Creating tree file at '{tree_file}'...")
        self._export_tree(tree_file)

        proof_file = os.path.join(output_folder, "all_proofs.json")
        logging.info(f"Creating proofs file at '{proof_file}'...")
        self._export_proofs(proof_file)

        parts_folder = os.path.join(output_folder, "parts")
        logging.info(f"Creating parts at '{parts_folder}/'...")
        self._write_parts(parts_folder)

    def get_parts(self):
        """Set parts appropiately

        Args:
            parts (list): List of parts to process

        Raises:
            pruvi.exceptions.NoPartsException.
        """
        if self._parts:
            return self._parts
        else:
            raise NoPartsException

    def set_parts(self, parts, codification="utf-8"):
        """Set parts appropiately

        Args:
            parts (list): List of parts to process.
            codification (str): The codification of the charset.
        """
        for p in parts:
            if isinstance(p, str):
                self._parts.append(p.encode(codification))
            elif isinstance(p, list) or isinstance(p, dict):
                element = json.dumps(p)
                self._parts.append(element.encode(codification))
            elif isinstance(p, bytes) or isinstance(p, bytearray):
                self._parts.append(p)
            else:
                self._parts.append(str(p).encode(codification))

    def verify_file(self, data_file, proof_file, merkle_root=None):
        """Verify a file

        Args:
            data_file (str): Path to the data file to verify.
            proof_file (str): Path to the proof file.
            merkle_root (str): A merkle root to proof.

        Returns:
            bool.
        """
        logging.info(f"Step 1/3: Hashing the file: '{data_file}'")
        with open(data_file, "rb") as f:
            data = f.read()
        calculated_digest = hash_bytes(data)
        logging.info(f"\tFile hash: '{calculated_digest}'")

        logging.info(f"Step 2/3: Verifying if the hash matches with that of the proof...")
        with open(proof_file) as f:
            proof_data = json.load(f)
        proof_index = proof_data["body"]["proof_index"]
        proof_digest = proof_data["body"]["proof_path"][proof_index][1]
        if proof_digest == calculated_digest:
            logging.info("\t✔️ OK!")
        else:
            logging.error("\t✕ Verification failed!")
            return

        logging.info(f"Step 3: Verifying the Merkle Proof...")
        if self.verify_proof(proof_data, merkle_root):
            logging.info("\t✔️ Verification successful!")
            return True
        else:
            logging.error("\t✕ Verification failed!")
            return False

    def verify_proof(self, data, merkle_root=None):
        """Verify a proof

        Args:
            data (dict): A Merkle proof to verify.
            merkle_root (str): A merkle root to proof.

        Returns:
            bool.
        """
        if not merkle_root:
            merkle_root = self.tree.rootHash
        else:
            if isinstance(merkle_root, str):
                merkle_root = merkle_root.encode()

        p = Proof(from_dict=data)
        return validateProof(target=merkle_root, proof=p)



from pruvi.splitters.file import FileSplitter


class AudioFileSplitter(FileSplitter):
    """Audio File Splitter object
    """
    def split_document(self, **kwargs):
        """Split an audio file

        Kwargs:
            seconds (int): Number of seconds in which to split the contents. Default: 1.
            format (str): Format of the output file. Default: "wav".
        """
        # Dirty hotfix to prevent pydub from printing information at the startup
        # This information will only be printed when the split_document is provided
        from pydub import AudioSegment
        self._parts = []
        audio = AudioSegment.from_wav(self.file_path)

        current_second = 0
        i = 0
        last_part = False

        while 1:
            end_second = current_second + kwargs.get("seconds", 1) * 1000
            if end_second > len(audio):
                end_second = len(audio)
                last_part = True

            new_part = audio[current_second:end_second]

            part_file = self._get_file_path(i)
            new_part.export(part_file, format=kwargs.get("format", "wav"))

            with open(part_file, "rb") as f:
                self._parts.append(f.read())

            if last_part:
                break
            else:
                current_second = end_second
                i += 1



from pruvi.splitters.base import BaseSplitter


class TextSplitter(BaseSplitter):
    """Base Splitter object
    """
    def split_document(self, text, method="lines"):
        """Split a text

        Args:
            text (str): Any kind of object to proof.
            method (str): The splitting method. One of the following: "lines", "paragraphs",
                "words".
        """
        parts = []

        if method == "lines":
            parts = text.splitlines()
        elif method == "words":
            parts = text.split()
        elif method == "paragraphs":
            parts = text.split("\n\n")
        else:
            parts = [text]

        self.set_parts(parts)



from pruvi.splitters.file import FileSplitter


class TextFileSplitter(FileSplitter):
    """Text File Splitter object
    """
    def split_document(self, **kwargs):
        """Split a binary file

        Kwargs:
            method (str): Splitting method. One of the following: "lines", "paragraphs",
                "words".
        """
        parts = []

        with open(self.file_path, 'r') as f:
            text = f.read()
            if kwargs.get("method") == "lines":
                parts = text.splitlines()
            elif kwargs.get("method") == "words":
                parts = text.split()
            elif kwargs.get("method") == "paragraphs":
                parts = text.split("\n\n")
            else:
                parts = [text]

        self.set_parts(parts)



from pruvi.splitters.file import FileSplitter


class BinaryFileSplitter(FileSplitter):
    """Binary File Splitter object
    """
    def split_document(self, **kwargs):
        """Split a binary file

        Kwargs:
            size (int): Number of bytes in which to split the contents
        """
        self._parts = []

        with open(self.file_path, 'rb') as f:
            while 1:
               content = f.read(kwargs.get("size"))
               if content:
                   self._parts.append(content)
               else:
                   break



from PyPDF2 import PdfFileWriter, PdfFileReader

from pruvi.splitters.file import FileSplitter


class PDFFileSplitter(FileSplitter):
    """PDF File Splitter object
    """
    def split_document(self, **kwargs):
        """Split a PDF file

        Kwargs:
            method (str): Splitting method. One of the following: "pages".
        """
        parts = []

        if kwargs.get("method", "pages") == "pages":
            inputpdf = PdfFileReader(open(self.file_path, "rb"))

            for i in range(inputpdf.numPages):
                output = PdfFileWriter()
                output.addPage(inputpdf.getPage(i))

                part_file = self._get_file_path(i)

                with open(part_file, "wb") as outputStream:
                    output.write(outputStream)

                with open(part_file, "rb") as f:
                    pdf_data = bytearray(f.read())
                parts.append(pdf_data)
        else:
            with open(self.file_path, 'rb') as f:
                parts = [f.read()]

        self.set_parts(parts)



import argparse
import json
import logging
import os
import sys

import pruvi
from pruvi.splitters.base import BaseSplitter
from pruvi.splitters.audio_file import AudioFileSplitter
from pruvi.splitters.binary_file import BinaryFileSplitter
from pruvi.splitters.pdf_file import PDFFileSplitter
from pruvi.splitters.text_file import TextFileSplitter

def main():
    """Main parser"""
    def get_parser():
        """Method that gets a valid parser

        Returns:
            argparse.ArgumentParser.
        """
        with open(
            os.path.join(
                os.path.dirname(pruvi.__file__),
                ".banner.txt"
            )
        ) as iF:
            banner = iF.read()

        print(banner)
        print("""
                            Coded with ♥ by @febrezo

        """)

        parser = argparse.ArgumentParser(
            description="Pruvi | A tool for generating partial proofs of big documents",
            epilog="For each subcommand, add '--help' for additional parameters.",
            add_help=False
        )

        about_parser = parser.add_argument_group(
            'Other commands',
            'Get additional information about this program.'
        )
        about_parser.add_argument(
            '-h', '--help',
            action='help',
            help='shows this help and exits.'
        )
        about_parser.add_argument(
            '--version',
            action='version',
            version='%(prog)s ' + pruvi.__version__,
            help='shows the version of this package and exits.'
        )
        about_parser.add_argument(
            '-L', '--log-level',
            metavar='<LOG_LEVEL>',
            action='store',
            default="INFO",
            choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
            help="The log level to be set."
        )

        # Add subparsers
        # --------------
        subparsers = parser.add_subparsers(
            description='Available commands.',
            dest='subcommand',
            metavar="<SUBCOMMAND>"
        )

        # Generate parser
        # --------------
        generate_subparser = subparsers.add_parser(
            'split',
            help='Split different files and obtain proofs.',
            conflict_handler='resolve'
        )

        generate_subparser.add_argument(
            '-o', '--output-folder',
            metavar='<OUTPUT_FOLDER>',
            action='store',
            default="./",
            help="The default output folder."
        )

        # Add parsers for from files
        # --------------------------
        from_subparsers = generate_subparser.add_subparsers(
            description='Available input files: json, csv.',
            dest='from_file',
            metavar="<FROM_FILE>"
        )

        # Audio subparser
        audio_subparser = from_subparsers.add_parser(
            'audio',
            help='get proofs from an audio file',
            conflict_handler='resolve'
        )
        audio_subparser.add_argument(
            '-f', '--file',
            metavar='<SOURCE_FILE>',
            required=True,
            help="the path to the given file to split."
        )
        audio_subparser.add_argument(
            '-s', '--seconds',
            metavar='<SECONDS>',
            action='store',
            type=int,
            required=True,
            help="The seconds in which the audio will be splitted."
        )

        # Binary subparser
        binary_subparser = from_subparsers.add_parser(
            'binary',
            help='get proofs from a binary file',
            conflict_handler='resolve'
        )
        binary_subparser.add_argument(
            '-f', '--file',
            metavar='<SOURCE_FILE>',
            required=True,
            help="the path to the given file to split."
        )
        binary_subparser.add_argument(
            '-b', '--bytes',
            metavar='<BYTE_NUMBER>',
            action='store',
            type=int,
            required=True,
            help="Number of bytes to use to split."
        )

        # PDF subparser
        pdf_subparser = from_subparsers.add_parser(
            'pdf',
            help='get proofs from the pages in a PDF file',
            conflict_handler='resolve'
        )
        pdf_subparser.add_argument(
            '-f', '--file',
            metavar='<SOURCE_FILE>',
            required=True,
            help="the path to the given file to split."
        )

        # Text file subparser
        text_file_subparser = from_subparsers.add_parser(
            'text',
            help='get proofs from a text file',
            conflict_handler='resolve'
        )
        text_file_subparser.add_argument(
            '-f', '--file',
            metavar='<SOURCE_FILE>',
            required=True,
            help="the path to the given file to split."
        )
        text_file_subparser.add_argument(
            '-m', '--method',
            metavar='<METHOD>',
            action='store',
            choices=["lines", "paragraphs", "words"],
            default="lines",
            help="Method to split the file."
        )

        # Subparser for validating proofs
        # -------------------------------
        validate_subparser = subparsers.add_parser(
            'validate',
            help='Validate a proof.',
        )

        template_main_subparser = validate_subparser.add_argument_group(
            'Validate a provided proof',
            'Verify if a given file and its proof are linked to a given Merkle root hash.'
        )
        template_main_subparser.add_argument(
            '-p', '--proof-file',
            metavar='<PROOF_FILE>',
            required=True,
            help="the path to a Merkle proof."
        )
        template_main_subparser.add_argument(
            '-d', '--data-file',
            metavar='<DATA_FILE>',
            required=True,
            help="the path to a data file."
        )
        template_main_subparser.add_argument(
            '-m', '--merkle-root',
            metavar='<MERKLE_ROOT>',
            required=True,
            help="the Merkle root that allegedly contains that file."
        )

        return parser

    parser = get_parser()

    args = parser.parse_args()

    logging.basicConfig(
        format='Pruvi [%(levelname)s] > %(message)s',
        level=logging.getLevelName(args.log_level)
    )

    if args.subcommand == "split":
        logging.info(f"Launch splitter process for '{args.from_file}' files...")

        if args.from_file == "audio":
            splitter = AudioFileSplitter(file_path=args.file)
            splitter.split_document(seconds=args.seconds)
        elif args.from_file == "binary":
            splitter = BinaryFileSplitter(file_path=args.file)
            splitter.split_document(size=args.bytes)
        elif args.from_file == "pdf":
            splitter = PDFFileSplitter(file_path=args.file)
            splitter.split_document()
        elif args.from_file == "text":
            splitter = TextFileSplitter(file_path=args.file)
            splitter.split_document(method=args.method)
        else:
            logging.error("No valid file type provided.")
            parser.print_help()
            sys.exit(1)

        logging.info("Creating tree...")
        splitter.create_tree()

        logging.info(f"Tree created. Merkle root hash: '{splitter.tree.rootHash.decode()}'")

        logging.info("Exporting proofs...")
        splitter.export(args.output_folder)
    elif args.subcommand == "validate":
        logging.info(f"Launch verification process for '{args.data_file}'...")
        splitter = BaseSplitter()
        splitter.verify_file(args.data_file, args.proof_file, args.merkle_root)
    else:
        logging.error(f"'{args.subcommand}' is not a valid subcommand.")
        parser.print_help()


if __name__ == "__main__":
    main()



###############################################################################
#
#   Copyright 2018-2019 Telefónica
#
#   csv2avro is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program. If not, see <http://www.gnu.org/licenses/>.
#
###############################################################################

from setuptools import find_packages, setup
import pruvi


setup(
    name='pruvi',
    version=pruvi.__version__,
    author='Félix Brezo (@febrezo)',
    description='Python3.6+ package to create proofs for certain parts of '
                'different types of documents using Merkle Trees.',
    url='https://github.com/febrezo/pruvi',
    include_package_data=True,
    license='GNU GPLv3+',
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'pruvi=pruvi.launcher:main',
        ],
    },
    install_requires=[
        'pydub',
        'pymerkle',
    ],
    dependency_links=[
        'git+git://github.com/febrezo/pymerkle/@pure-byte-hashing#egg=pymerkle'
    ],
)



import hashlib


def hash_bytes(data, hash_type="sha3_512"):
    """Hash a sequence of bytes

    Args:
        data (bytes): Set of hashes to hash.

    Returns:
        str. Hex digest.

    Raises:
        AttributeError.
    """
    m = getattr(hashlib, hash_type)()
    m.update(data)
    return m.hexdigest()



import os
import tempfile

from pruvi.splitters.base import BaseSplitter


class FileSplitter(BaseSplitter):
    """File Splitter object

    Attributes:
        file_path (str): The file to split.
    """
    def __init__(self, file_path, parts=[], hash_type="sha3_512"):
        """Constructor

        Args:
            file_path (str): The file to split.
            parts (list): List of elements to proof.
            hash_type (str): The hash type to be used.
        """
        BaseSplitter.__init__(self, parts=parts, hash_type=hash_type)
        self.file_path = file_path

    def _get_file_path(self, part):
        return os.path.join(
            tempfile.gettempdir(),
            f"document-part{part}"
        )

    def split_document(self, text, method="lines"):
        """Split a text

        Args:
            text (str): Any kind of object to proof.
            method (str): The splitting method.

        Returns:
            bytes.
        """
        self._parts = []

        if method == "lines":
            self._parts = text.splitlines()
        elif method == "sentences":
            self._parts = text.split(".")
        else:
            self._parts = [text]
